'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var path = require('path');
var fsExtra = require('fs-extra');
var graphql = require('graphql');
var graphqlConfig = require('graphql-config');
var graphqlConfigUtilities = require('graphql-config-utilities');
var graphqlToolUtilities = require('graphql-tool-utilities');
var validate = require('./validate.js');

async function evaluateFixtures(fixturePaths, {
  cwd
}) {
  const config = await graphqlConfig.loadConfig({
    rootDir: path.resolve(cwd)
  });
  if (!config) {
    return [];
  }
  const projectASTCollection = await Promise.all(graphqlConfigUtilities.getGraphQLProjects(config).map(getOperationsForProject));
  return runForEachFixture(fixturePaths, fixture => evaluateFixture(fixture, projectASTCollection));
}
async function getOperationsForProject(projectConfig) {
  const operationPaths = await graphqlConfigUtilities.getGraphQLProjectIncludedFilePaths(projectConfig);
  const operationSources = await Promise.all(operationPaths.map(loadOperationSource));
  const document = graphql.concatAST(operationSources.map(source => {
    try {
      return graphql.parse(source);
    } catch (error) {
      throw new Error(`Error parsing '${source.name}':\n\n${error.message.replace(/Syntax Error.*?\(.*?\) /, '')}`);
    }
  }));
  let schema;
  try {
    schema = await projectConfig.getSchema();
  } catch (error) {
    throw new Error(`Error parsing '${projectConfig.schema}':\n\n${error.message.replace(/Syntax Error.*?\(.*?\) /, '')}`);
  }
  return {
    ast: graphqlToolUtilities.compile(schema, document),
    config: projectConfig
  };
}
async function loadOperationSource(filePath) {
  const body = await fsExtra.readFile(filePath, 'utf8');
  return new graphql.Source(body, filePath);
}
function runForEachFixture(fixturePaths, runner) {
  return Promise.all(fixturePaths.map(async fixturePath => {
    try {
      const fixture = await fsExtra.readJSON(fixturePath);
      return {
        fixturePath,
        ...runner({
          path: fixturePath,
          content: fixture
        })
      };
    } catch (error) {
      return {
        fixturePath,
        scriptError: error,
        validationErrors: []
      };
    }
  }));
}
function evaluateFixture(fixture, projectASTCollection) {
  try {
    const {
      operation,
      projectAST: {
        ast
      }
    } = validate.getOperationForFixture(fixture, projectASTCollection);
    return validate.validateFixture(fixture, ast, operation);
  } catch (error) {
    if (error instanceof validate.MissingOperationError || error instanceof validate.AmbiguousOperationNameError) {
      return {
        fixturePath: fixture.path,
        validationErrors: [error]
      };
    }
    throw error;
  }
}

exports.evaluateFixtures = evaluateFixtures;
