import { resolve } from 'path';
import { readFile, readJSON } from 'fs-extra';
import { concatAST, parse, Source } from 'graphql';
import { loadConfig } from 'graphql-config';
import { getGraphQLProjects, getGraphQLProjectIncludedFilePaths } from 'graphql-config-utilities';
import { compile } from 'graphql-tool-utilities';
import { getOperationForFixture, validateFixture, MissingOperationError, AmbiguousOperationNameError } from './validate.mjs';

async function evaluateFixtures(fixturePaths, {
  cwd
}) {
  const config = await loadConfig({
    rootDir: resolve(cwd)
  });
  if (!config) {
    return [];
  }
  const projectASTCollection = await Promise.all(getGraphQLProjects(config).map(getOperationsForProject));
  return runForEachFixture(fixturePaths, fixture => evaluateFixture(fixture, projectASTCollection));
}
async function getOperationsForProject(projectConfig) {
  const operationPaths = await getGraphQLProjectIncludedFilePaths(projectConfig);
  const operationSources = await Promise.all(operationPaths.map(loadOperationSource));
  const document = concatAST(operationSources.map(source => {
    try {
      return parse(source);
    } catch (error) {
      throw new Error(`Error parsing '${source.name}':\n\n${error.message.replace(/Syntax Error.*?\(.*?\) /, '')}`);
    }
  }));
  let schema;
  try {
    schema = await projectConfig.getSchema();
  } catch (error) {
    throw new Error(`Error parsing '${projectConfig.schema}':\n\n${error.message.replace(/Syntax Error.*?\(.*?\) /, '')}`);
  }
  return {
    ast: compile(schema, document),
    config: projectConfig
  };
}
async function loadOperationSource(filePath) {
  const body = await readFile(filePath, 'utf8');
  return new Source(body, filePath);
}
function runForEachFixture(fixturePaths, runner) {
  return Promise.all(fixturePaths.map(async fixturePath => {
    try {
      const fixture = await readJSON(fixturePath);
      return {
        fixturePath,
        ...runner({
          path: fixturePath,
          content: fixture
        })
      };
    } catch (error) {
      return {
        fixturePath,
        scriptError: error,
        validationErrors: []
      };
    }
  }));
}
function evaluateFixture(fixture, projectASTCollection) {
  try {
    const {
      operation,
      projectAST: {
        ast
      }
    } = getOperationForFixture(fixture, projectASTCollection);
    return validateFixture(fixture, ast, operation);
  } catch (error) {
    if (error instanceof MissingOperationError || error instanceof AmbiguousOperationNameError) {
      return {
        fixturePath: fixture.path,
        validationErrors: [error]
      };
    }
    throw error;
  }
}

export { evaluateFixtures };
