import { basename, dirname } from 'path';
import { isNonNullType, isListType, isObjectType, GraphQLString, GraphQLInt, GraphQLFloat, GraphQLBoolean, isScalarType, isEnumType } from 'graphql';

class MissingOperationError extends Error {
  constructor({
    path
  }, lookForOperationNames, projectASTCollection) {
    super([`Could not find a matching operation for '${path}'`, `(looked for ${lookForOperationNames.join(', ')}).`, `Make sure to put your fixture in a folder named the same as the operation,`, `or add an '${OPERATION_MARKER}' key indicating the operation.`, `Available operations: ${projectASTCollection.flatMap(({
      ast: {
        operations
      }
    }) => Object.keys(operations)).join(', ')}`].join(' '));
  }
}
class AmbiguousOperationNameError extends Error {
  constructor({
    path
  }, foundOperations) {
    super([`Ambiguous operation name found for '${path}'`, `(found ${Array.from(new Set(foundOperations.map(({
      lookForOperationName
    }) => lookForOperationName))).join(', ')})`, `in projects:`, `${foundOperations.map(({
      projectAST: {
        config
      }
    }) => config.name).join(', ')}.`, `Try renaming the operation in one of the projects listed and updating`, `the fixture folder name or use an '${OPERATION_MARKER}' key indicating`, `the new operation name.`].join(' '));
  }
}
const OPERATION_MARKER = '@operation';
function normalizeOperationName(operationName) {
  return operationName ? operationName.replace(/(Query|Mutation|Subscription)$/i, '') : undefined;
}
function getOperationNames(fixture) {
  const fixtureDirectoryName = basename(dirname(fixture.path));
  const operationMarkerName = fixture.content[OPERATION_MARKER];
  return Array.from(new Set([fixtureDirectoryName, operationMarkerName, normalizeOperationName(fixtureDirectoryName), normalizeOperationName(operationMarkerName)])).filter(operationName => typeof operationName === 'string');
}
function findOperations(lookForOperationNames, projectASTCollection) {
  return projectASTCollection.map(projectAST => {
    for (const lookForOperationName of lookForOperationNames) {
      const operation = projectAST.ast.operations[lookForOperationName];
      if (operation) {
        return {
          lookForOperationName,
          operation,
          projectAST
        };
      }
    }
    return null;
  }).filter(match => Boolean(match));
}
function getOperationForFixture(fixture, projectASTCollection) {
  const lookForOperationNames = getOperationNames(fixture);
  const operations = findOperations(lookForOperationNames, projectASTCollection);
  if (operations.length === 0) {
    throw new MissingOperationError(fixture, lookForOperationNames, projectASTCollection);
  }
  if (operations.length > 1) {
    throw new AmbiguousOperationNameError(fixture, operations);
  }
  return operations[0];
}
function validateFixture(fixture, ast, operation) {
  const {
    fields = [],
    filePath,
    operationType
  } = operation;
  const value = {
    ...fixture.content
  };
  delete value[OPERATION_MARKER];
  return {
    fixturePath: fixture.path,
    operationName: operation.operationName,
    operationType,
    operationPath: filePath === 'GraphQL request' ? undefined : filePath,
    validationErrors: fields.reduce((allErrors, field) => {
      return allErrors.concat(validateValueAgainstFieldDescription(value[field.responseName], field, '', ast));
    }, [])
  };
}
function validateValueAgainstFieldDescription(value, fieldDescription, parentKeyPath, ast) {
  const {
    type,
    responseName
  } = fieldDescription;
  const keyPath = updateKeyPath(parentKeyPath, responseName);
  const typeErrors = validateValueAgainstType(value, type, keyPath, {
    shallow: true
  });
  if (typeErrors.length > 0) {
    return typeErrors;
  }
  return Array.isArray(value) ? validateListAgainstFieldDescription(value, fieldDescription, type, keyPath, ast) : validateValueAgainstObjectFieldDescription(value, fieldDescription, keyPath, ast);
}
function validateValueAgainstObjectFieldDescription(value, fieldDescription, keyPath, ast) {
  if (value == null || typeof value !== 'object' || Array.isArray(value)) {
    return [];
  }
  const {
    fields = [],
    fragmentSpreads = [],
    type
  } = fieldDescription;
  const fragmentFields = [];
  if (fragmentSpreads) {
    fragmentSpreads.map(spread => ast.fragments[spread]).forEach(fragment => {
      fragment.fields.forEach(field => {
        if (fields.some(({
          fieldName
        }) => fieldName === field.fieldName)) {
          return;
        }
        const isGuaranteedTypeMatch = fragment.possibleTypes.includes(makeTypeNullable(type));
        fragmentFields.push(isGuaranteedTypeMatch ? field : {
          ...field,
          type: makeTypeNullable(field.type)
        });
      });
    });
  }
  return validateValueAgainstFields(value, fields.concat(fragmentFields), keyPath, ast);
}
function makeTypeNullable(type) {
  return isNonNullType(type) ? type.ofType : type;
}
function validateListAgainstFieldDescription(value, fieldDescription, type, keyPath, ast) {
  const itemType = isNonNullType(type) ? type.ofType.ofType : type.ofType;
  return value.reduce((allErrors, item, index) => {
    const itemKeyPath = updateKeyPath(keyPath, index);
    const itemTypeErrors = validateValueAgainstType(item, itemType, itemKeyPath, {
      shallow: true
    });
    if (itemTypeErrors.length > 0) {
      return allErrors.concat(itemTypeErrors);
    }
    return Array.isArray(item) ? allErrors.concat(validateListAgainstFieldDescription(item, fieldDescription, itemType, itemKeyPath, ast)) : allErrors.concat(validateValueAgainstObjectFieldDescription(item, fieldDescription, itemKeyPath, ast));
  }, []);
}
function validateValueAgainstFields(value, fields, keyPath, ast) {
  const finalValue = value || {};
  const excessFields = Object.keys(finalValue).filter(key => fields.every(({
    responseName
  }) => key !== responseName)).map(key => error(keyPath, `has key '${key}' not present in the query`));
  return fields.reduce((allErrors, field) => {
    return allErrors.concat(validateValueAgainstFieldDescription(finalValue[field.responseName], field, keyPath, ast));
  }, excessFields);
}
function validateValueAgainstType(value, type, keyPath, options = {
  shallow: false
}) {
  const {
    shallow
  } = options;
  if (isNonNullType(type)) {
    return value == null ? [error(keyPath, `should be non-null but was ${String(value)}`)] : validateValueAgainstType(value, type.ofType, keyPath, options);
  }
  if (value === null) {
    return [];
  }
  const valueType = typeof value;
  if (isListType(type)) {
    if (!Array.isArray(value)) {
      return [error(keyPath, `should be an array (or null), but was ${articleForType(valueType)} ${valueType}`)];
    }
    return shallow ? [] : value.reduce((allErrors, item, index) => allErrors.concat(validateValueAgainstType(item, type.ofType, updateKeyPath(keyPath, index), options)), []);
  }
  if (value === undefined) {
    const typeName = nameForType(type);
    return [error(keyPath, `should be ${articleForType(typeName)} ${typeName} (or null), but was undefined`)];
  }
  if (isObjectType(type)) {
    if (valueType === 'object') {
      if (shallow) {
        return [];
      } else {
        const fields = type.getFields();
        return Object.keys(value).reduce((fieldErrors, key) => {
          const fieldKeyPath = updateKeyPath(keyPath, key);
          return fields[key] == null ? fieldErrors.concat([error(fieldKeyPath, `does not exist on type ${type.name} (available fields: ${Object.keys(fields).join(', ')})`)]) : fieldErrors.concat(validateValueAgainstType(value[key], fields[key].type, fieldKeyPath, options));
        }, []);
      }
    } else {
      return [error(keyPath, `should be an object but was a ${valueType}`)];
    }
  }
  if (type === GraphQLString) {
    return valueType === 'string' ? [] : [error(keyPath, `should be a string but was a ${valueType}`)];
  } else if (type === GraphQLInt) {
    if (typeof value === 'number') {
      return Number.isInteger(value) ? [] : [error(keyPath, 'should be an integer but was a float')];
    }
    return [error(keyPath, `should be an integer but was a ${valueType}`)];
  } else if (type === GraphQLFloat) {
    return Number.isNaN(value) ? [error(keyPath, `should be a float but was a ${valueType}`)] : [];
  } else if (type === GraphQLBoolean) {
    return typeof value === 'boolean' ? [] : [error(keyPath, `should be a boolean but was a ${valueType}`)];
  }
  if (isScalarType(type)) {
    return type.parseValue(value) == null ? [error(keyPath, `value does not match scalar ${nameForType(type)}`)] : [];
  }
  if (isEnumType(type)) {
    try {
      return type.parseValue(value) == null ? [error(keyPath, `value does not match enum ${nameForType(type)} (available values: ${type.getValues().map(enumValue => enumValue.value).join(', ')})`)] : [];
    } catch (err) {
      return [error(keyPath, err.message)];
    }
  }
  return [];
}
const CUSTOM_NAMES = {
  [GraphQLBoolean.name]: 'boolean',
  [GraphQLFloat.name]: 'float',
  [GraphQLInt.name]: 'integer',
  [GraphQLString.name]: 'string'
};
function nameForType(type) {
  return Object.prototype.hasOwnProperty.call(CUSTOM_NAMES, type.name) ? CUSTOM_NAMES[type.name] : type.name;
}
const TYPES_WITH_ARTICLE_AN = ['object', 'integer', 'array'];
function articleForType(type) {
  return TYPES_WITH_ARTICLE_AN.includes(type) ? 'an' : 'a';
}
function updateKeyPath(keyPath, newKey) {
  if (typeof newKey === 'number') {
    return `${keyPath}[${newKey}]`;
  }
  return keyPath ? `${keyPath}.${newKey}` : newKey;
}
function error(keyPath, message) {
  return {
    keyPath,
    message
  };
}

export { AmbiguousOperationNameError, MissingOperationError, findOperations, getOperationForFixture, getOperationNames, validateFixture };
